@@ -418,6 +418,18 @@ void wmf_ipa_font_init (wmfAPI* API,wmfAPI_Options* options)
 		{	wmf_ipa_font_map_xml (API,&(font_data->FD),WMF_XTRA_FONTMAP);
 		}
 	}
+
+	font_data->GS.max = 0;
+	font_data->GS.len = 0;
+
+	font_data->GS.FI = 0;
+
+	if (API->flags & WMF_OPT_GS_FONTMAP)
+	{	wmf_ipa_font_map_gs (API,&(font_data->GS),options->gs_fontmap_file);
+	}
+	else
+	{	wmf_ipa_font_map_gs (API,&(font_data->GS),WMF_GS_FONTMAP);
+	}
 }
 
 void wmf_ipa_font_map_set (wmfAPI* API,wmfMap map)
@@ -426,358 +438,6 @@ void wmf_ipa_font_map_set (wmfAPI* API,wmfMap map)
 	font_data->map = map;
 }
 
-static wmfXML_FontInfo* ipa_font_sys_map (wmfAPI* API,wmfFont* font)
-{	wmfFontData* font_data = (wmfFontData*) API->font_data;
-
-	wmfXML_FontInfo* FI = 0;
-
-	unsigned int i;
-
-	for (i = 0; i < font_data->FD.len; i++)
-	{	if (strcmp (font->lfFaceName,font_data->FD.FI[i].familyname) == 0)
-		{	if ( wmf_strstr (font_data->FD.FI[i].fullname,"Italic")
-			  || wmf_strstr (font_data->FD.FI[i].fullname,"Oblique"))
-			{	if (font->lfItalic != 1) continue;
-			}
-			else
-			{	if (font->lfItalic == 1) continue;
-			}
-			if ( wmf_strstr (font_data->FD.FI[i].weight,"Bold")
-			  || wmf_strstr (font_data->FD.FI[i].weight,"Demi")) /* or "Medium" ?? */
-			{	if (font->lfWeight <= 550) continue;
-			}
-			else
-			{	if (font->lfWeight >  550) continue;
-			}
-			FI = &(font_data->FD.FI[i]);
-			break;
-		}
-	}
-
-	return (FI);
-}
-
-static FT_Face ipa_font_sys_face (wmfAPI* API,wmfFont* font,wmfXML_FontInfo* FI)
-{	wmfFontData* font_data = (wmfFontData*) API->font_data;
-
-	wmfFT_CacheEntry entry;
-
-	FT_Face face = 0;
-
-	unsigned int i;
-
-	struct stat stat_buf;
-
-	if (FI == 0) return (0);
-
-	if (FI->glyphs == 0)
-	{	WMF_DEBUG (API,"no glyphs in system-font record?");
-		return (0);
-	}
-
-	face = 0;
-
-	/* Check cache for font
-	 */
-	i = 0;
-	while (font_data->cache[i].name)
-	{	if (strcmp (font_data->cache[i].path,FI->glyphs) == 0)
-		{	face = font_data->cache[i].face;
-			break;
-		}
-		i++;
-	}
-
-	/* If found, set font and return
-	 */
-	if (face)
-	{	font->ps_name = FI->name;
-		font->ft_face = face;
-		return (face);
-	}
-
-	if (stat (FI->glyphs,&stat_buf))
-	{	WMF_ERROR (API,"unable to stat font file:");
-		WMF_ERROR (API,FI->glyphs);
-		API->err = wmf_E_BadFile;
-		return (0);
-	}
-
-	if (FT_New_Face (font_data->Library,FI->glyphs,0,&face) != 0)
-	{	WMF_ERROR (API,"Failed to open font:");
-		WMF_ERROR (API,FI->glyphs);
-		API->err = wmf_E_DeviceError;
-		return (0);
-	}
-
-	if (FI->metrics)
-	{	if (stat (FI->metrics,&stat_buf))
-		{	WMF_DEBUG (API,"unable to stat font metrics file:");
-			WMF_DEBUG (API,FI->metrics);
-		}
-		else
-		{	if (FT_Attach_File (face,FI->metrics) != 0)
-			{	WMF_DEBUG (API,"unable to load font metrics file:");
-				WMF_DEBUG (API,FI->metrics);
-			}
-		}
-	}
-
-	/* Select encoding (I'm very uncertain about this!): - [TODO] ??
-	 */
-	if (FT_Select_Charmap (face,ft_encoding_adobe_standard) == 0)
-	{	WMF_DEBUG (API,FI->fullname);
-		WMF_DEBUG (API,"Adobe Standard Encoding");
-	}
-	else if (FT_Select_Charmap (face,ft_encoding_adobe_custom) == 0)
-	{	WMF_DEBUG (API,FI->fullname);
-		WMF_DEBUG (API,"Adobe Custom Encoding");
-	}
-	else if (FT_Select_Charmap (face,ft_encoding_symbol) == 0)
-	{	WMF_DEBUG (API,FI->fullname);
-		WMF_DEBUG (API,"Symbol Encoding");
-	}
-	else if (FT_Select_Charmap (face,ft_encoding_unicode) == 0)
-	{	WMF_DEBUG (API,FI->fullname);
-		WMF_DEBUG (API,"Unicode Encoding");
-	}
-	else
-	{	WMF_ERROR (API,"Bad encoding! (Please help me!)");
-		API->err = wmf_E_DeviceError;
-	}
-
-	entry.name = FI->name;
-	entry.path = FI->glyphs;
-
-	entry.face = face;
-
-	ipa_font_add_cache (API,&entry); /* cache font name, path & face */
-
-	ipa_font_add_api (API,FI->name); /* add font name to list of fonts used */
-
-	font->ps_name = FI->name;
-	font->ft_face = face;
-
-	return (face);
-}
-
-void wmf_ipa_font_map (wmfAPI* API,wmfFont* font)
-{	wmfFontData* font_data = (wmfFontData*) API->font_data;
-
-	wmfFT_Mapping* font_mapping;
-
-	wmfFT_CacheEntry entry;
-
-	char* mapping = 0;
-
-	unsigned int i;
-	unsigned int index = 0;
-	unsigned int count;
-	unsigned int length;
-
-	struct stat stat_buf;
-
-	font->ps_name = 0;
-	font->ft_face = 0;
-
-/* Check system fonts for match and load font face if found...
- */
-	if (ipa_font_sys_face (API,font,ipa_font_sys_map (API,font))) return;
-
-/* otherwise, load one of standard 13...
- */
-	mapping = 0;
-
-	/* First: check for an exact match:
-	 */
-	i = 0;
-	while (font_data->wmf[i].name)
-	{	if (strcmp (font->lfFaceName,font_data->wmf[i].name) == 0)
-		{	mapping = font_data->wmf[i].name;
-			break;
-		}
-		i++;
-	}
-
-	/* If no exact match, check for a close (i.e., sub-string) match:
-	 */
-	if (mapping == 0)
-	{	i = 0;
-		while (font_data->sub[i].name)
-		{	if (wmf_strstr (font->lfFaceName,font_data->sub[i].name))
-			{	mapping = font_data->sub[i].mapping;
-				break;
-			}
-			i++;
-		}
-	}
-
-	/* If still no match, use the default:
-	 */
-	if (mapping == 0) mapping = DefaultFontMapping;
-
-	/* Check chosen mapping in list of exact names:
-	 */
-	i = 0;
-	while (font_data->wmf[i].name)
-	{	if (strcmp (mapping,font_data->wmf[i].name) == 0)
-		{	index = i;
-			break;
-		}
-		i++;
-	}
-
-	/* If not found then something odd has happened - probably a typo somewhere...
-	 */
-	if (font_data->wmf[i].name == 0)
-	{	WMF_ERROR (API,"Glitch! Unmapped font...");
-		API->err = wmf_E_Glitch;
-		return;
-	}
-
-	/* Select a new mapping (the PS font name) on basis of weight & italic parameters
-	 */
-	if (font->lfWeight > 550) /* => Bold (?? Don't ask me...) */
-	{	if (font->lfItalic == 1) /* Italic */
-		{	mapping = font_data->wmf[index].bolditalic;
-		}
-		else
-		{	mapping = font_data->wmf[index].bold;
-		}
-	}
-	else
-	{	if (font->lfItalic == 1) /* Italic */
-		{	mapping = font_data->wmf[index].italic;
-		}
-		else
-		{	mapping = font_data->wmf[index].normal;
-		}
-	}
-
-	font_mapping = 0;
-
-	/* Get list member with filename info etc.
-	 */
-	i = 0;
-	while (font_data->ps[i].name)
-	{	if (strcmp (mapping,font_data->ps[i].name) == 0)
-		{	font_mapping = &(font_data->ps[i]);
-			break;
-		}
-		i++;
-	}
-
-	/* If not found then something odd has happened - probably a typo somewhere...
-	 */
-	if (font_mapping == 0)
-	{	WMF_ERROR (API,"Glitch! Unmapped font...");
-		API->err = wmf_E_Glitch;
-		return;
-	}
-
-	if (font_mapping->face) /* Have face already... */
-	{	font->ps_name = font_mapping->name;
-		font->ft_face = font_mapping->face;
-		return;
-	}
-
-	ipa_font_add_api (API,font_mapping->name); /* add font name to list of fonts used */
-
-	if (ERR (API))
-	{	WMF_DEBUG (API,"bailing...");
-		return;
-	}
-
-	/* Determine length of string required to hold path name of font
-	 */
-	count = 0;
-	i = 0;
-	while (font_data->fontdirs[i])
-	{	length = strlen (font_data->fontdirs[i]);
-		if (count < length) count = length;
-		i++;
-	}
-	count += 1 + strlen (font_mapping->mapping) + 1;
-
-	mapping = (char*) wmf_malloc (API,count);
-
-	if (ERR (API))
-	{	WMF_DEBUG (API,"bailing...");
-		return;
-	}
-
-	/* Compose full font file path:
-	 */
-	i = 0;
-	while (font_data->fontdirs[i])
-	{	strcpy (mapping,font_data->fontdirs[i]);
-		strcat (mapping,"/");
-		strcat (mapping,font_mapping->mapping);
-		WMF_DEBUG (API,mapping);
-		if (stat (mapping,&stat_buf) == 0) break; /* file exists */
-		i++;
-	}
-
-	/* Open font face if file found:
-	 */
-	if (font_data->fontdirs[i] == 0)
-	{	WMF_ERROR (API,"Unable to locate font file:");
-		WMF_ERROR (API,font_mapping->mapping);
-		API->err = wmf_E_BadFile;
-	}
-	else if (FT_New_Face (font_data->Library,mapping,0,&(font_mapping->face)) != 0)
-	{	WMF_ERROR (API,"Failed to open font:");
-		WMF_ERROR (API,mapping);
-		API->err = wmf_E_DeviceError;
-		font_mapping->face = 0;
-	}
-
-	if (ERR (API))
-	{	WMF_DEBUG (API,"bailing...");
-		return;
-	}
-
-	entry.name = font_mapping->name;
-	entry.face = font_mapping->face;
-
-	entry.path = wmf_strdup (API,mapping); /* caching performed at end; see below */
-
-	if (ERR (API))
-	{	WMF_DEBUG (API,"bailing...");
-		return;
-	}
-
-	/* If PFA or PFB look for associated metrics AFM file:
-	 */
-	if ((strcmp (mapping + strlen (mapping) - 4,".pfa") == 0)
-	 || (strcmp (mapping + strlen (mapping) - 4,".pfb") == 0))
-	{	mapping[strlen (mapping) - 4] = 0;
-
-		strcat (mapping,".afm");
-
-		if (FT_Attach_File (font_mapping->face,mapping) != 0)
-		{	WMF_ERROR (API,"Failed to open metrics:");
-			WMF_ERROR (API,mapping);
-			API->err = wmf_E_DeviceError;
-		}
-	}
-
-	/* Finally, select encoding:
-	 */
-	if (FT_Select_Charmap (font_mapping->face,font_mapping->encoding))
-	{	WMF_ERROR (API,"Bad encoding!");
-		API->err = wmf_E_DeviceError;
-	}
-
-	ipa_font_add_cache (API,&entry); /* cache font name, path & face */
-
-	wmf_free (API,mapping);
-	wmf_free (API,entry.path);
-
-	font->ps_name = font_mapping->name;
-	font->ft_face = font_mapping->face;
-}
-
 void wmf_ipa_font_dir (wmfAPI* API,char* fontdir)
 {	wmfFontData* font_data = (wmfFontData*) API->font_data;
 
@@ -1221,6 +881,11 @@ static void xml2_start (void* user_data,const char* tag,const char** attributes)
 	FI.version = 0;
 	FI.alias = 0;
 
+	if (attributes == 0)
+	{	WMF_DEBUG (API,"xml2_start: found NULL instead of ptr-to-NULL!\n");
+		return;
+	}
+
 	attr = attributes;
 	while (*attr)
 	{	if (strcmp ((*attr),"format") == 0)
@@ -1281,7 +946,7 @@ static void xml2fontmap_read (wmfAPI* API,wmfXML_FontData* FD,char* xmlfontmap)
 
 	xmlParserCtxtPtr ctxt;
 
-	xmlSAXHandler sax = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
+	xmlSAXHandler sax = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
 
 	sax.startElement = (startElementSAXFunc) xml2_start;
 
@@ -1303,3 +968,865 @@ static void xml2fontmap_read (wmfAPI* API,wmfXML_FontData* FD,char* xmlfontmap)
 }
 
 #endif /* HAVE_XML2 */
+
+void wmf_ipa_font_map_gs (wmfAPI* API,wmfGS_FontData* FD,char* gsfontmap)
+{	ipa_font_gs_file (API,FD,gsfontmap);
+
+	if (ERR (API))
+	{	WMF_DEBUG (API,"bailing...");
+		return;
+	}
+
+	/* TODO ?? */
+}
+
+static void ipa_font_gs_file (wmfAPI* API,wmfGS_FontData* FD,char* file)
+{	char* line = 0;
+
+	char* name = 0;
+	char* alias = 0;
+
+	FILE* in = 0;
+
+	in = fopen (file,"r");
+
+	if (in == 0)
+	{	WMF_DEBUG (API,"ipa_font_gs_file: unable to read ghostscript fontmap!\n");
+		return;
+	}
+
+	while (wmf_true)
+	{	line = ipa_font_gs_readline (API,in);
+
+		if (line == 0) break;
+
+		if (*line)
+		{	name = line;
+			alias = line;
+
+			while (!isspace ((int) (*alias)))
+			{	if ((*alias) == 0) break;
+				alias++;
+			}
+			if (*alias)
+			{	(*alias) = 0;
+				alias++;
+			}
+
+			ipa_font_gs_add (API,FD,name,alias);
+		}
+
+		wmf_free (API,line);
+	}
+
+	fclose (in);
+}
+
+static void ipa_font_gs_add (wmfAPI* API,wmfGS_FontData* FD,char* name,char* alias)
+{	wmfGS_FontInfo* more = 0;
+
+	wmf_bool_t fMatched;
+
+	unsigned int i;
+
+	if ((name == 0) || (alias == 0)) return;
+
+	if (name[0] != '/') return;
+
+	if ((alias[0] != '/') && (alias[0] != '(')) return;
+
+	if ((alias[0] == '(') && (alias[strlen (alias) - 1] != ')')) return;
+
+	if (FD->FI == 0)
+	{	FD->FI = (wmfGS_FontInfo*) wmf_calloc (API,0x20,sizeof (wmfGS_FontInfo));
+
+		if (ERR (API))
+		{	WMF_DEBUG (API,"bailing...");
+			return;
+		}
+
+		FD->max = 0x20;
+		FD->len = 0;
+	}
+
+	fMatched = wmf_false;
+	for (i = 0; i < FD->len; i++)
+	{	if (strcmp (FD->FI[i].name,name+1) == 0)
+		{	wmf_free (API,FD->FI[i].alias);
+			FD->FI[i].alias = wmf_strdup (API,alias);
+			fMatched = wmf_true;
+			break;
+		}
+	}
+	if (fMatched) return;
+
+	if (FD->len == FD->max)
+	{	more = (wmfGS_FontInfo*) wmf_realloc (API,FD->FI,(FD->max+0x20) * sizeof (wmfGS_FontInfo));
+
+		if (ERR (API))
+		{	WMF_DEBUG (API,"bailing...");
+			return;
+		}
+
+		FD->FI = more;
+
+		FD->max += 0x20;
+	}
+
+	FD->FI[FD->len].name  = wmf_strdup (API,name+1);
+	FD->FI[FD->len].alias = wmf_strdup (API,alias);
+
+	FD->len++;
+}
+
+static char* ipa_font_gs_alias (wmfAPI* API,wmfGS_FontData* FD,char* name)
+{	char* alias = 0;
+
+	unsigned int i;
+
+	for (i = 0; i < FD->len; i++)
+	{	if (strcmp (FD->FI[i].name,name) == 0)
+		{	alias = FD->FI[i].alias;
+			break;
+		}
+	}
+
+	return (alias);
+}
+
+static char* ipa_font_gs_readline (wmfAPI* API,FILE* in)
+{	char buf[128];
+
+	char* line = 0;
+	char* more = 0;
+	char* ptr = 0;
+	char* ptr1 = 0;
+	char* ptr2 = 0;
+
+	wmf_bool_t fBackSlash;
+	wmf_bool_t fContinue;
+	wmf_bool_t fReadExtra = wmf_false;
+
+	if (fgets (buf,128,in) == 0) return (0);
+
+	fBackSlash = wmf_false;
+	fContinue  = wmf_true;
+	ptr = buf;
+	while (*ptr)
+	{	if (((*ptr) == '%') && (!fBackSlash))
+		{	fContinue = wmf_false;
+			if (buf[strlen(buf)-1] != '\n')
+			{	fReadExtra = wmf_true;
+			}
+			else
+			{	fReadExtra = wmf_false;
+			}
+			(*ptr) = 0;
+			break;
+		}
+		if ((*ptr) == '\n')
+		{	if (fBackSlash)
+			{	fContinue = wmf_true;
+				ptr--;
+			}
+			else
+			{	fContinue = wmf_false;
+			}
+			(*ptr) = 0;
+			break;
+		}
+		if ((*ptr) == '\\')
+		{	fBackSlash = wmf_true;
+		}
+		else
+		{	fBackSlash = wmf_false;
+		}
+		ptr++;
+	}
+
+	line = wmf_strdup (API,buf);
+
+	if (line == 0) return (line);
+
+	while (fContinue)
+	{	if (fgets (buf,10,in) == 0)
+		{	wmf_free (API,line);
+			return (0);
+		}
+
+		fBackSlash = wmf_false;
+		fContinue  = wmf_true;
+		ptr = buf;
+		while (*ptr)
+		{	if (((*ptr) == '%') && (!fBackSlash))
+			{	fContinue = wmf_false;
+				fReadExtra = wmf_true;
+				(*ptr) = 0;
+				break;
+			}
+			if ((*ptr) == '\n')
+			{	if (fBackSlash)
+				{	fContinue = wmf_true;
+					ptr--;
+				}
+				else
+				{	fContinue = wmf_false;
+				}
+				(*ptr) = 0;
+				break;
+			}
+			if ((*ptr) == '\\')
+			{	fBackSlash = wmf_true;
+			}
+			else
+			{	fBackSlash = wmf_false;
+			}
+			ptr++;
+		}
+
+		more = wmf_str_append (API,line,buf);
+
+		wmf_free (API,line);
+
+		line = more;
+
+		if (line == 0)
+		{	fContinue = wmf_false;
+			break;
+		}
+	}
+
+	if (line == 0) return (0);
+
+	if (fReadExtra) while (buf[strlen(buf)-1] != '\n') fgets (buf,128,in);
+
+	/* Strip the string */
+
+	ptr1 = line;
+	ptr2 = line;
+	while (*ptr2)
+	{	if (!isspace ((int) (*ptr2))) break;
+		ptr2++;
+	}
+	while (*ptr2)
+	{	if (isspace ((int) (*ptr2))) break;
+		(*ptr1) = (*ptr2);
+		ptr1++;
+		ptr2++;
+	}
+	if (*ptr2)
+	{	(*ptr1) = (*ptr2);
+		ptr1++;
+		ptr2++;
+	}
+	while (*ptr2)
+	{	if (!isspace ((int) (*ptr2))) break;
+		ptr2++;
+	}
+	while (*ptr2)
+	{	if (isspace ((int) (*ptr2))) break;
+		(*ptr1) = (*ptr2);
+		ptr1++;
+		ptr2++;
+	}
+	(*ptr1) = 0;
+
+	while ((--ptr1) >= line)
+	{	if (!isspace ((int) (*ptr1))) break;
+		(*ptr1) = 0;
+	}
+
+	ptr1 = line;
+	ptr2 = line;
+	while (*ptr2)
+	{	if ((*ptr2) == '%') ptr1--;
+		(*ptr1) = (*ptr2);
+		ptr1++;
+		ptr2++;
+	}
+	(*ptr1) = 0;
+
+	return (line);
+}
+
+/**
+ * This is the fun bit.
+ * 
+ * @param API  the API handle
+ * @param font WMF font object: font's \b family \b name, \b italics flag, and the \b weight
+ * 
+ * Other inputs include: a ghostscript-style Fontmap, and up to two XML RedHat-style fontmaps. (These are
+ * all parsed earlier.) There are also the fallback 13 standard postscript fonts. And then there is a list
+ * of font directories to search for fonts.
+ * 
+ * \b libwmf requires fonts to have a postscript name, and maintains an internal cache of glyphs path and
+ * \b freetype font face, keyed to postscript font name. Unfortunately, although the XML fontmaps do have
+ * font \b family \b name, the GS fontmap does not. In the latter, therefore, matching WMF font info to
+ * postscript font name is hit-and-miss.
+ */
+void wmf_ipa_font_map (wmfAPI* API,wmfFont* font)
+{	char* mapping = 0;
+
+	font->ps_name = 0;
+	font->ft_face = 0;
+
+/* Check system fonts for match and load font face if found...
+ */
+	if (ipa_font_sys_face (API,font,ipa_font_sys_map (API,font))) return;
+
+/* Check GS fontmap for match and load font face if found...
+ */
+	if (ipa_font_gs_face (API,font,ipa_font_gs_map (API,font))) return;
+
+/* otherwise, load one of standard 13...
+ */
+	mapping = ipa_font_std (API,font);
+
+	if (mapping) ipa_font_face (API,font,mapping);
+}
+
+/**
+ * Searches XML Fontmap for font-name matching WMF's font name.
+ * 
+ * Basically, this is matching a font's family name + italic + weight parameters to a postscript name.
+ * Font weight is problematic.
+ */
+static wmfXML_FontInfo* ipa_font_sys_map (wmfAPI* API,wmfFont* font)
+{	wmfFontData* font_data = (wmfFontData*) API->font_data;
+
+	wmfXML_FontInfo* FI = 0;
+
+	unsigned int i;
+
+	for (i = 0; i < font_data->FD.len; i++)
+	{	if (strcmp (font->lfFaceName,font_data->FD.FI[i].familyname) == 0)
+		{	if ( wmf_strstr (font_data->FD.FI[i].fullname,"Italic")
+			  || wmf_strstr (font_data->FD.FI[i].fullname,"Oblique"))
+			{	if (font->lfItalic != 1) continue;
+			}
+			else
+			{	if (font->lfItalic == 1) continue;
+			}
+			if ( wmf_strstr (font_data->FD.FI[i].weight,"Bold")
+			  || wmf_strstr (font_data->FD.FI[i].weight,"Demi")) /* or "Medium" ?? */
+			{	if (font->lfWeight <= 550) continue;
+			}
+			else
+			{	if (font->lfWeight >  550) continue;
+			}
+			FI = &(font_data->FD.FI[i]);
+			break;
+		}
+	}
+
+	return (FI);
+}
+
+/**
+ * Checks XML Fontmap entry is valid before loading face.
+ */
+static FT_Face ipa_font_sys_face (wmfAPI* API,wmfFont* font,wmfXML_FontInfo* FI)
+{	wmfFontData* font_data = (wmfFontData*) API->font_data;
+
+	FT_Face face = 0;
+
+	struct stat stat_buf;
+
+	if (FI == 0) return (0);
+
+	if (FI->name == 0)
+	{	WMF_DEBUG (API,"no postscript name in system-font record?");
+		return (0);
+	}
+
+	if (FI->glyphs == 0)
+	{	WMF_DEBUG (API,"no glyphs in system-font record?");
+		return (0);
+	}
+
+	face = ipa_font_face (API,font,FI->name);
+
+	return (face);
+}
+
+/**
+ * Searches GS Fontmap for font-name matching WMF's font name.
+ * 
+ * Basically, this is matching a font's family name + italic + weight parameters to a postscript name.
+ * I have decided to test only the first word of the family name, so incorrect matches are probably.
+ * Font weight is problematic as well.
+ */
+static wmfGS_FontInfo* ipa_font_gs_map (wmfAPI* API,wmfFont* font)
+{	wmfFontData* font_data = (wmfFontData*) API->font_data;
+
+	wmfGS_FontData* GS = &(font_data->GS);
+	wmfGS_FontInfo* FI = 0;
+
+	wmf_bool_t require_italic;
+	wmf_bool_t require_bold;
+	wmf_bool_t found_italic;
+	wmf_bool_t found_bold;
+
+	unsigned int name_length = 0;
+	unsigned int i;
+
+	char* name = 0;
+	char* ptr = 0;
+
+	if (GS->len == 0) return (0);
+
+	name = font->lfFaceName;
+	if (name == 0)
+	{	WMF_DEBUG (API,"No font name?");
+		API->err = wmf_E_Glitch;
+		return (0);
+	}
+
+	/* Find first white-space character or eol
+	 */
+	ptr = name;
+	while (!isspace (*ptr))
+	{	if ((*ptr) == 0) break;
+		ptr++;
+		name_length++;
+	}
+
+	if (name_length == 0)
+	{	WMF_DEBUG (API,"Unexpectedly short font name?");
+		API->err = wmf_E_Glitch;
+		return (0);
+	}
+
+	require_italic = ((font->lfItalic == 1) ? wmf_true : wmf_false);
+
+	require_bold = ((font->lfWeight > 550) ? wmf_true : wmf_false);
+
+	/* Search for match in GS Fontmap list
+	 */
+	for (i = 0; i < GS->len; i++)
+	{	if (strncmp (name,GS->FI[i].name,name_length) == 0)
+		{	found_italic = wmf_false;
+			found_bold   = wmf_false;
+			if ( wmf_strstr (GS->FI[i].name,"Ital")
+			  || wmf_strstr (GS->FI[i].name,"Obli"))
+			{	found_italic = wmf_true;
+			}
+			if ( wmf_strstr (GS->FI[i].name,"Bold")
+			  || wmf_strstr (GS->FI[i].name,"Demi")) /* or "Medi" ?? */
+			{	found_bold   = wmf_true;
+			}
+			if ((require_italic == found_italic) && (require_bold == found_bold))
+			{	FI = &(GS->FI[i]);
+				break;
+			}
+		}
+	}
+
+	if (FI == 0) return (0);
+
+	return (FI);
+}
+
+/**
+ * Takes a [name]/[alias] pair from the GS Fontmap and tries to load [name] as a font,
+ * and then, provided [alias] is aliased to a font-file & not a 3rd font-name, [alias].
+ */
+static FT_Face ipa_font_gs_face (wmfAPI* API,wmfFont* font,wmfGS_FontInfo* FI)
+{	wmfFontData* font_data = (wmfFontData*) API->font_data;
+
+	FT_Face face = 0;
+
+	unsigned int i;
+
+	char* aalias = 0;
+
+	if (FI == 0) return (0);
+
+	face = ipa_font_face (API,font,FI->name);
+
+	if (face) return (face);
+
+	if (FI->alias[0] != '/') return (0);
+
+	aalias = ipa_font_gs_alias (API,&(font_data->GS),FI->alias+1);
+
+	if (aalias[0] != '(')
+	{	WMF_DEBUG (API,"font lookup is too complicated! Giving up...");
+		return (0);
+	}
+
+	face = ipa_font_face (API,font,FI->alias+1);
+
+	return (face);
+}
+
+/**
+ * Takes a postscript font name and:
+ * 
+ * (1) Check to see whether font has been cached; if so: return (ipa_font_face_cached ())
+ * (2) Check for font in XML table; if found:
+ *    (a) ipa_font_face_open ()
+ *    (b) return (ipa_font_face_cached ())
+ * (3) Check for font with file in GS table; if so:
+ *    (a) search for font file in font search path; if found:
+ *       ( i) ipa_font_face_open ()
+ *       (ii) return (ipa_font_face_cached ())
+ * (4) Check for font in internal table; if found:
+ *    (a) ipa_font_face_open ()
+ *    (b) return (ipa_font_face_cached ())
+ * (5) return (0)
+ */
+static FT_Face ipa_font_face (wmfAPI* API,wmfFont* font,char* ps_name)
+{	wmfFontData* font_data = (wmfFontData*) API->font_data;
+
+	FT_Face face = 0;
+
+	unsigned int i;
+
+	char* name = 0;
+	char* suffix = 0;
+	char* glyphs = 0;
+	char* metrics = 0;
+
+	/* (1) Check to see whether font has been cached; if so: return (ipa_font_face_cached ())
+	 */
+	face = ipa_font_face_cached (API,font,ps_name);
+
+	if (face) return (face);
+
+	/* (2) Check for font in XML table
+	 */
+	glyphs = 0;
+	metrics = 0;
+	for (i = 0; i < font_data->FD.len; i++)
+	{	if (font_data->FD.FI[i].name == 0) continue;
+		if (strcmp (ps_name,font_data->FD.FI[i].name) == 0)
+		{	glyphs  = font_data->FD.FI[i].glyphs;
+			metrics = font_data->FD.FI[i].metrics;
+			break;
+		}
+	}
+
+	if (glyphs)
+	{	face = ipa_font_face_open (API,ps_name,glyphs,metrics);
+
+		if (face) return (ipa_font_face_cached (API,font,ps_name));
+	}
+
+	/* (3) Check for font with file in GS table
+	 */
+	name = 0;
+	for (i = 0; i < font_data->GS.len; i++)
+	{	if (strcmp (ps_name,font_data->GS.FI[i].name) == 0)
+		{	if (font_data->GS.FI[i].alias[0] == '(')
+			{	name = wmf_strdup (API,font_data->GS.FI[i].alias + 1);
+			}
+			break;
+		}
+	}
+
+	glyphs = 0;
+	if (name)
+	{	name[strlen (name) - 1] = 0;
+
+		glyphs = ipa_font_path_find (API,name);
+	}
+
+	metrics = 0;
+	if (glyphs)
+	{	if (strlen (name) > 3)
+		{	suffix = name + strlen (name) - 4;
+			if ((strcmp (suffix,".pfa") == 0) || (strcmp (suffix,".pfb") == 0))
+			{	strcpy (suffix,".afm");
+				metrics = ipa_font_path_find (API,name);
+			}
+		}
+
+		face = ipa_font_face_open (API,ps_name,glyphs,metrics);
+	}
+
+	if (name) wmf_free (API,name);
+
+	if (glyphs)  wmf_free (API,glyphs);
+	if (metrics) wmf_free (API,metrics);
+
+	if (face) return (ipa_font_face_cached (API,font,ps_name));
+
+	/* (4) Check for font in internal table
+	 */
+	name = 0;
+	i = 0;
+	while (font_data->ps[i].name)
+	{	if (strcmp (ps_name,font_data->ps[i].name) == 0)
+		{	name = wmf_strdup (API,font_data->ps[i].mapping);
+			break;
+		}
+		i++;
+	}
+
+	glyphs = 0;
+	if (name)
+	{	glyphs = ipa_font_path_find (API,name);
+	}
+
+	metrics = 0;
+	if (glyphs)
+	{	if (strlen (name) > 3)
+		{	suffix = name + strlen (name) - 4;
+			if ((strcmp (suffix,".pfa") == 0) || (strcmp (suffix,".pfb") == 0))
+			{	strcpy (suffix,".afm");
+				metrics = ipa_font_path_find (API,name);
+			}
+		}
+
+		face = ipa_font_face_open (API,ps_name,glyphs,metrics);
+	}
+
+	if (name) wmf_free (API,name);
+
+	if (glyphs)  wmf_free (API,glyphs);
+	if (metrics) wmf_free (API,metrics);
+
+	if (face) return (ipa_font_face_cached (API,font,ps_name));
+
+	/* (5) return (0)
+	 */
+	return (0);
+}
+
+/**
+ * Opens the font with freetype and caches postscript name, glyphs path & FT font face
+ */
+static FT_Face ipa_font_face_open (wmfAPI* API,char* ps_name,char* glyphs,char* metrics)
+{	wmfFontData* font_data = (wmfFontData*) API->font_data;
+
+	wmfFT_CacheEntry entry;
+
+	FT_Face face = 0;
+
+	struct stat stat_buf;
+
+	if (stat (glyphs,&stat_buf))
+	{	WMF_ERROR (API,"unable to stat font file:");
+		WMF_ERROR (API,glyphs);
+		API->err = wmf_E_BadFile;
+		return (0);
+	}
+
+	if (FT_New_Face (font_data->Library,glyphs,0,&face) != 0)
+	{	WMF_ERROR (API,"Failed to open font:");
+		WMF_ERROR (API,glyphs);
+		API->err = wmf_E_DeviceError;
+		return (0);
+	}
+
+	if (metrics)
+	{	if (stat (metrics,&stat_buf))
+		{	WMF_DEBUG (API,"unable to stat font metrics file:");
+			WMF_DEBUG (API,metrics);
+		}
+		else
+		{	if (FT_Attach_File (face,metrics) != 0)
+			{	WMF_DEBUG (API,"unable to load font metrics file:");
+				WMF_DEBUG (API,metrics);
+			}
+		}
+	}
+
+	/* Select encoding (I'm very uncertain about this!): - [TODO] ??
+	 */
+	if (FT_Select_Charmap (face,ft_encoding_adobe_standard) == 0)
+	{	WMF_DEBUG (API,glyphs);
+		WMF_DEBUG (API,"Adobe Standard Encoding");
+	}
+	else if (FT_Select_Charmap (face,ft_encoding_adobe_custom) == 0)
+	{	WMF_DEBUG (API,glyphs);
+		WMF_DEBUG (API,"Adobe Custom Encoding");
+	}
+	else if (FT_Select_Charmap (face,ft_encoding_symbol) == 0)
+	{	WMF_DEBUG (API,glyphs);
+		WMF_DEBUG (API,"Symbol Encoding");
+	}
+	else if (FT_Select_Charmap (face,ft_encoding_unicode) == 0)
+	{	WMF_DEBUG (API,glyphs);
+		WMF_DEBUG (API,"Unicode Encoding");
+	}
+	else
+	{	WMF_ERROR (API,"Bad encoding! (Please help me!)");
+		API->err = wmf_E_DeviceError;
+	}
+
+	entry.name = ps_name;
+	entry.path = glyphs;
+
+	entry.face = face;
+
+	ipa_font_add_cache (API,&entry); /* cache font name, path & face */
+
+	ipa_font_add_api (API,ps_name); /* add font name to list of fonts used */
+
+	return (face);
+}
+
+/**
+ * Finds font info in cache and sets WMF's font entry appropriately.
+ */
+static FT_Face ipa_font_face_cached (wmfAPI* API,wmfFont* font,char* ps_name)
+{	wmfFontData* font_data = (wmfFontData*) API->font_data;
+
+	FT_Face face = 0;
+
+	unsigned int i;
+
+	/* Check cache for font
+	 */
+	i = 0;
+	while (font_data->cache[i].name)
+	{	if (strcmp (font_data->cache[i].name,ps_name) == 0)
+		{	face = font_data->cache[i].face;
+			break;
+		}
+		i++;
+	}
+
+	if (face)
+	{	font->ps_name = ps_name;
+		font->ft_face = face;
+	}
+
+	return (face);
+}
+
+/**
+ * Search for file_name in font path and return full path to file as wmf_malloc() string
+ */
+static char* ipa_font_path_find (wmfAPI* API,char* file_name)
+{	wmfFontData* font_data = (wmfFontData*) API->font_data;
+
+	struct stat stat_buf;
+
+	unsigned int length;
+	unsigned int count;
+	unsigned int i;
+
+	char* path = 0;
+
+	/* Determine length of string required to hold path name of font
+	 */
+	count = 0;
+	i = 0;
+	while (font_data->fontdirs[i])
+	{	length = strlen (font_data->fontdirs[i]);
+		if (count < length) count = length;
+		i++;
+	}
+	count += 1 + strlen (file_name) + 1;
+
+	path = (char*) wmf_malloc (API,count);
+
+	if (ERR (API))
+	{	WMF_DEBUG (API,"bailing...");
+		return (0);
+	}
+
+	/* Compose full font file path:
+	 */
+	i = 0;
+	while (font_data->fontdirs[i])
+	{	strcpy (path,font_data->fontdirs[i]);
+		strcat (path,"/");
+		strcat (path,file_name);
+		WMF_DEBUG (API,path);
+		if (stat (path,&stat_buf) == 0) break; /* file exists */
+		i++;
+	}
+
+	if (font_data->fontdirs[i] == 0)
+	{	WMF_DEBUG (API,"file not found in font path.");
+		wmf_free (API,path);
+		path = 0;
+	}
+
+	return (path);
+}
+
+/**
+ * If after XML & GS fontmap searches the font still hasn't been matched, then use one of the 13 standard
+ * postscript fonts (or any others that have been added via library options).
+ */
+static char* ipa_font_std (wmfAPI* API,wmfFont* font)
+{	wmfFontData* font_data = (wmfFontData*) API->font_data;
+
+	char* mapping = 0;
+
+	unsigned int i;
+	unsigned int index = 0;
+
+	/* First: check for an exact match:
+	 */
+	i = 0;
+	while (font_data->wmf[i].name)
+	{	if (strcmp (font->lfFaceName,font_data->wmf[i].name) == 0)
+		{	mapping = font_data->wmf[i].name;
+			break;
+		}
+		i++;
+	}
+
+	/* If no exact match, check for a close (i.e., sub-string) match:
+	 */
+	if (mapping == 0)
+	{	i = 0;
+		while (font_data->sub[i].name)
+		{	if (wmf_strstr (font->lfFaceName,font_data->sub[i].name))
+			{	mapping = font_data->sub[i].mapping;
+				break;
+			}
+			i++;
+		}
+	}
+
+	/* If still no match, use the default:
+	 */
+	if (mapping == 0) mapping = DefaultFontMapping;
+
+	/* Check chosen mapping in list of exact names:
+	 */
+	i = 0;
+	while (font_data->wmf[i].name)
+	{	if (strcmp (mapping,font_data->wmf[i].name) == 0)
+		{	index = i;
+			break;
+		}
+		i++;
+	}
+
+	/* If not found then something odd has happened - probably a typo somewhere...
+	 */
+	if (font_data->wmf[i].name == 0)
+	{	WMF_ERROR (API,"Glitch! Unmapped font...");
+		API->err = wmf_E_Glitch;
+		return (0);
+	}
+
+	/* Select a new mapping (the PS font name) on basis of weight & italic parameters
+	 */
+	if (font->lfWeight > 550) /* => Bold (?? Don't ask me...) */
+	{	if (font->lfItalic == 1) /* Italic */
+		{	mapping = font_data->wmf[index].bolditalic;
+		}
+		else
+		{	mapping = font_data->wmf[index].bold;
+		}
+	}
+	else
+	{	if (font->lfItalic == 1) /* Italic */
+		{	mapping = font_data->wmf[index].italic;
+		}
+		else
+		{	mapping = font_data->wmf[index].normal;
+		}
+	}
+
+	return (mapping);
+}
