@@ -283,6 +283,8 @@ static void meta_arc (wmfAPI* API,wmfRecord* Record)
 
 	wmfDrawArc_t drawarc;
 
+	wmfPen* pen = 0;
+
 	U16 par_U16_x;
 	U16 par_U16_y;
 	U16 end_x;
@@ -334,7 +336,9 @@ static void meta_arc (wmfAPI* API,wmfRecord* Record)
 	drawarc.start = wmf_D_Coord_translate (API,l_pt);
 
 	if (SCAN (API))
-	{	scope = (float) (MAX (P->dc->pen->width,P->dc->pen->height)) / 2;
+	{	pen = WMF_DC_PEN (P->dc);
+
+		scope = (float) (MAX (WMF_PEN_WIDTH (pen),WMF_PEN_HEIGHT (pen))) / 2;
 
 		D_Coord_Register (API,drawarc.end,scope);
 		D_Coord_Register (API,drawarc.start,scope);
@@ -540,6 +544,8 @@ static void meta_ellipse (wmfAPI* API,wmfRecord* Record)
 
 	wmfDrawArc_t drawarc;
 
+	wmfPen* pen = 0;
+
 	U16 par_U16_x;
 	U16 par_U16_y;
 
@@ -565,7 +571,10 @@ static void meta_ellipse (wmfAPI* API,wmfRecord* Record)
 	drawarc.BR = wmf_D_Coord_translate (API,l_pt);
 
 	if (SCAN (API))
-	{	scope = (float) (MAX (P->dc->pen->width,P->dc->pen->height)) / 2;
+	{	pen = WMF_DC_PEN (P->dc);
+
+		scope = (float) (MAX (WMF_PEN_WIDTH (pen),WMF_PEN_HEIGHT (pen))) / 2;
+
 		D_Coord_Register (API,drawarc.TL,scope);
 		D_Coord_Register (API,drawarc.BR,scope);
 		return;
@@ -584,6 +593,8 @@ static void meta_line (wmfAPI* API,wmfRecord* Record)
 
 	wmfDrawLine_t drawline;
 
+	wmfPen* pen = 0;
+
 	U16 par_U16_x;
 	U16 par_U16_y;
 
@@ -606,7 +617,10 @@ static void meta_line (wmfAPI* API,wmfRecord* Record)
 	P->current = l_pt;
 
 	if (SCAN (API))
-	{	scope = (float) (MAX (P->dc->pen->width,P->dc->pen->height)) / 2;
+	{	pen = WMF_DC_PEN (P->dc);
+
+		scope = (float) (MAX (WMF_PEN_WIDTH (pen),WMF_PEN_HEIGHT (pen))) / 2;
+
 		D_Coord_Register (API,drawline.from,scope);
 		D_Coord_Register (API,drawline.to,scope);
 		return;
@@ -626,6 +640,8 @@ static void meta_lines (wmfAPI* API,wmfRecord* Record)
 
 	wmfPolyLine_t polyline;
 
+	wmfPen* pen = 0;
+
 	U16 par_U16_x;
 	U16 par_U16_y;
 
@@ -647,7 +663,10 @@ static void meta_lines (wmfAPI* API,wmfRecord* Record)
 	}
 
 	if (SCAN (API))
-	{	scope = (float) (MAX (P->dc->pen->width,P->dc->pen->height)) / 2;
+	{	pen = WMF_DC_PEN (P->dc);
+
+		scope = (float) (MAX (WMF_PEN_WIDTH (pen),WMF_PEN_HEIGHT (pen))) / 2;
+
 		index = 1;
 		for (i = 0; i < polyline.count; i++)
 		{	par_U16_x = ParU16 (API,Record,index);
@@ -694,6 +713,8 @@ static void meta_polygon (wmfAPI* API,wmfRecord* Record)
 
 	wmfPolyLine_t polyline;
 
+	wmfPen* pen = 0;
+
 	U16 par_U16_x;
 	U16 par_U16_y;
 
@@ -715,7 +736,10 @@ static void meta_polygon (wmfAPI* API,wmfRecord* Record)
 	}
 
 	if (SCAN (API))
-	{	scope = (float) (MAX (P->dc->pen->width,P->dc->pen->height)) / 2;
+	{	pen = WMF_DC_PEN (P->dc);
+
+		scope = (float) (MAX (WMF_PEN_WIDTH (pen),WMF_PEN_HEIGHT (pen))) / 2;
+
 		index = 1;
 		for (i = 0; i < polyline.count; i++)
 		{	par_U16_x = ParU16 (API,Record,index);
@@ -764,6 +788,8 @@ static void meta_polygons (wmfAPI* API,wmfRecord* Record)
 
 	wmfRecord Polygon;
 
+	wmfPen* pen = 0;
+
 	U16 par_U16_x;
 	U16 par_U16_y;
 
@@ -808,7 +834,10 @@ static void meta_polygons (wmfAPI* API,wmfRecord* Record)
 	}
 
 	if (SCAN (API))
-	{	scope = (float) (MAX (P->dc->pen->width,P->dc->pen->height)) / 2;
+	{	pen = WMF_DC_PEN (P->dc);
+
+		scope = (float) (MAX (WMF_PEN_WIDTH (pen),WMF_PEN_HEIGHT (pen))) / 2;
+
 		index = 1 + num_poly;
 		for (i = 0; i < num_pars; i++)
 		{	par_U16_x = ParU16 (API,Record,index);
@@ -861,6 +890,8 @@ static void meta_round (wmfAPI* API,wmfRecord* Record)
 
 	wmfDrawRectangle_t drawrect;
 
+	wmfPen* pen = 0;
+
 	U16 par_U16_x;
 	U16 par_U16_y;
 
@@ -886,7 +917,10 @@ static void meta_round (wmfAPI* API,wmfRecord* Record)
 	drawrect.BR = wmf_D_Coord_translate (API,l_pt);
 
 	if (SCAN (API))
-	{	scope = (float) (MAX (P->dc->pen->width,P->dc->pen->height)) / 2;
+	{	pen = WMF_DC_PEN (P->dc);
+
+		scope = (float) (MAX (WMF_PEN_WIDTH (pen),WMF_PEN_HEIGHT (pen))) / 2;
+
 		D_Coord_Register (API,drawrect.TL,scope);
 		D_Coord_Register (API,drawrect.BR,scope);
 		return;
@@ -911,6 +945,8 @@ static void meta_rect (wmfAPI* API,wmfRecord* Record)
 
 	wmfDrawRectangle_t drawrect;
 
+	wmfPen* pen = 0;
+
 	U16 par_U16_x;
 	U16 par_U16_y;
 
@@ -936,7 +972,10 @@ static void meta_rect (wmfAPI* API,wmfRecord* Record)
 	drawrect.BR = wmf_D_Coord_translate (API,l_pt);
 
 	if (SCAN (API))
-	{	scope = (float) (MAX (P->dc->pen->width,P->dc->pen->height)) / 2;
+	{	pen = WMF_DC_PEN (P->dc);
+
+		scope = (float) (MAX (WMF_PEN_WIDTH (pen),WMF_PEN_HEIGHT (pen))) / 2;
+
 		D_Coord_Register (API,drawrect.TL,scope);
 		D_Coord_Register (API,drawrect.BR,scope);
 		return;
@@ -1007,6 +1046,11 @@ static void meta_rgn_brush (wmfAPI* API,wmfRecord* Record)
 	obj_region = objects + oid_region;
 	obj_brush  = objects + oid_brush;
 
+	if (SCAN (API) && DIAG (API))
+	{	diagnose_object (API,(unsigned int) oid_region,obj_region);
+		diagnose_object (API,(unsigned int) oid_brush, obj_brush );
+	}
+
 	if ((obj_region->type != OBJ_REGION) || (obj_brush->type != OBJ_BRUSH))
 	{	WMF_ERROR (API,"libwmf: have lost track of the objects in this metafile");
 		WMF_ERROR (API,"        please send it to us at http://www.wvware.com/");
@@ -1089,25 +1133,25 @@ static void meta_rgn_brush (wmfAPI* API,wmfRecord* Record)
 	{
 	case META_FRAMEREGION:
 		if (FR->region_frame)
-		{	temp_brush = P->dc->brush;
+		{	temp_brush = WMF_DC_BRUSH (polyrect.dc); /* ultimately redundant ?? */
 
-			P->dc->brush = brush;
+			WMF_DC_SET_BRUSH (polyrect.dc,brush);
 
 			FR->region_frame (API,&polyrect);
 
-			P->dc->brush = temp_brush;
+			WMF_DC_SET_BRUSH (polyrect.dc,temp_brush); /* ultimately redundant ?? */
 		}
 	break;
 
 	case META_FILLREGION:
 		if (FR->region_paint)
-		{	temp_brush = P->dc->brush;
+		{	temp_brush = WMF_DC_BRUSH (polyrect.dc); /* ultimately redundant ?? */
 
-			P->dc->brush = brush;
+			WMF_DC_SET_BRUSH (polyrect.dc,brush);
 
 			FR->region_paint (API,&polyrect);
 
-			P->dc->brush = temp_brush;
+			WMF_DC_SET_BRUSH (polyrect.dc,temp_brush); /* ultimately redundant ?? */
 		}
 	break;
 
@@ -1170,6 +1214,10 @@ static void meta_rgn_paint (wmfAPI* API,wmfRecord* Record)
 
 	obj_region = objects + oid_region;
 
+	if (SCAN (API) && DIAG (API))
+	{	diagnose_object (API,(unsigned int) oid_region,obj_region);
+	}
+
 	if (obj_region->type != OBJ_REGION)
 	{	WMF_ERROR (API,"libwmf: have lost track of the objects in this metafile");
 		WMF_ERROR (API,"        please send it to us at http://www.wvware.com/");
@@ -1232,13 +1280,13 @@ static void meta_rgn_paint (wmfAPI* API,wmfRecord* Record)
 	{
 	case META_INVERTREGION:
 		if (FR->region_paint)
-		{	temp_rop = P->dc->ROPmode;
+		{	temp_rop = WMF_DC_ROP (polyrect.dc); /* ultimately redundant ?? */
 
-			P->dc->ROPmode = R2_NOT;
+			WMF_DC_SET_ROP (polyrect.dc,R2_NOT);
 
 			FR->region_paint (API,&polyrect);
 
-			P->dc->ROPmode = temp_rop;
+			WMF_DC_SET_ROP (polyrect.dc,temp_rop); /* ultimately redundant ?? */
 		}
 	break;
 
@@ -1326,6 +1374,8 @@ static void meta_rgn_create (wmfAPI* API,wmfRecord* Record)
 	if (SCAN (API) && DIAG (API))
 	{	fprintf (stderr,"\t[0x%04x]",Record->function);
 		fprintf (stderr,"\t#par=%lu; index 0-4,6-10 skipped; max. index = 5",Record->size);
+
+		diagnose_object (API,(unsigned int) oid_region,obj_region);
 	}
 
 	num_band = ParU16 (API,Record,5);
@@ -1420,10 +1470,26 @@ static void meta_clip_select (wmfAPI* API,wmfRecord* Record)
 
 	obj_region = objects + oid_region;
 
+	if (SCAN (API) && DIAG (API))
+	{	diagnose_object (API,(unsigned int) oid_region,obj_region);
+	}
+
 	if (obj_region->type != OBJ_REGION)
-	{	WMF_ERROR (API,"libwmf: have lost track of the objects in this metafile");
-		WMF_ERROR (API,"        please send it to us at http://www.wvware.com/");
-		API->err = wmf_E_Glitch;
+	{	if (API->flags & WMF_OPT_IGNORE_NONFATAL)
+		{	/* Some metafiles use this even though no region-objects have been defined,
+			 * so I have some doubt about the correctness of this handler; perhaps
+			 * this should reset the clip region to the entire space?
+			 * 
+			 * Anyway, I am making this particular error non-fatal.
+			 */
+			WMF_DEBUG (API,"libwmf: have lost track of the objects in this metafile");
+			WMF_DEBUG (API,"        please send it to us at http://www.wvware.com/");
+		}
+		else
+		{	WMF_ERROR (API,"libwmf: have lost track of the objects in this metafile");
+			WMF_ERROR (API,"        please send it to us at http://www.wvware.com/");
+			API->err = wmf_E_Glitch;
+		}
 		return;
 	}
 
@@ -1855,6 +1921,8 @@ static void meta_dib_brush (wmfAPI* API,wmfRecord* Record)
 	if (SCAN (API) && DIAG (API))
 	{	fprintf (stderr,"\t[0x%04x]",Record->function);
 		fprintf (stderr,"\t#par=%lu; max. index = ?",Record->size);
+
+		diagnose_object (API,(unsigned int) oid_brush,obj_brush);
 	}
 
 	bmp_record = OffsetRecord (API,Record,2);
@@ -1890,13 +1958,15 @@ static void meta_dib_brush (wmfAPI* API,wmfRecord* Record)
 		return;
 	}
 
-	brush->lbStyle = BS_DIBPATTERN;
-	brush->lbColor = wmf_rgb_white (API);
-	brush->bmp = bmp_read.bmp;
+	WMF_BRUSH_SET_STYLE (brush,BS_DIBPATTERN);
 
-	if (SCAN (API)) wmf_ipa_color_add (API,&(brush->lbColor));
+	WMF_BRUSH_SET_COLOR (brush,&wmf_black);
 
-	P->dc->brush = brush;
+	WMF_BRUSH_SET_BITMAP (brush,&(bmp_read.bmp));
+
+	if (SCAN (API)) wmf_ipa_color_add (API,WMF_BRUSH_COLOR (brush));
+
+	WMF_DC_SET_BRUSH (P->dc,brush);
 }
 
 static void meta_rop_draw (wmfAPI* API,wmfRecord* Record)
@@ -2005,27 +2075,33 @@ static void meta_dc_set (wmfAPI* API,wmfRecord* Record)
 	switch (Record->function)
 	{
 	case META_SETROP2:
-		P->dc->ROPmode = par_U16;
+		WMF_DC_SET_ROP (P->dc,par_U16);
 	break;
 
 	case META_SETTEXTJUSTIFICATION:
-		P->dc->breakextra = par_U16;
+		WMF_DC_SET_BREAKEXTRA (P->dc,par_U16);
 	break;
 
 	case META_SETTEXTCHAREXTRA:
-		P->dc->charextra = par_U16;
+		WMF_DC_SET_CHAREXTRA (P->dc,par_U16);
 	break;
 
 	case META_SETPOLYFILLMODE:
-		P->dc->polyfillmode = par_U16;
+		WMF_DC_SET_POLYFILL (P->dc,par_U16);
 	break;
 
 	case META_SETTEXTALIGN:
-		P->dc->textalign = par_U16;
+		WMF_DC_SET_TEXTALIGN (P->dc,par_U16);
 	break;
 
 	case META_SETBKMODE:
-		P->dc->bgmode = par_U16;
+		if (par_U16 == TRANSPARENT)
+		{	WMF_DC_SET_TRANSPARENT (P->dc);
+		}
+		else
+		{	WMF_DC_SET_OPAQUE (P->dc);
+			if (par_U16 != OPAQUE) WMF_DEBUG (API,"unexpected background mode; assuming opaque...");
+		}
 	break;
 
 	default:
@@ -2059,11 +2135,11 @@ static void meta_dc_color (wmfAPI* API,wmfRecord* Record)
 	switch (Record->function)
 	{
 	case META_SETTEXTCOLOR:
-		P->dc->textcolor = color;
+		WMF_DC_SET_TEXTCOLOR (P->dc,&color);
 	break;
 
 	case META_SETBKCOLOR:
-		P->dc->bgcolor = color;
+		WMF_DC_SET_BACKGROUND (P->dc,&color);
 	break;
 
 	default:
@@ -2099,18 +2175,22 @@ static void meta_dc_select (wmfAPI* API,wmfRecord* Record)
 
 	obj = objects + oid;
 
+	if (SCAN (API) && DIAG (API))
+	{	diagnose_object (API,(unsigned int) oid,obj);
+	}
+
 	switch (obj->type)
 	{
+	case OBJ_BRUSH:
+		WMF_DC_SET_BRUSH (P->dc,&(obj->obj.brush));
+	break;
+
 	case OBJ_PEN:
-		P->dc->pen = &(obj->obj.pen);
+		WMF_DC_SET_PEN (P->dc,&(obj->obj.pen));
 	break;
 
 	case OBJ_FONT:
-		P->dc->font = &(obj->obj.font);
-	break;
-
-	case OBJ_BRUSH:
-		P->dc->brush = &(obj->obj.brush);
+		WMF_DC_SET_FONT (P->dc,&(obj->obj.font));
 	break;
 
 	default:
@@ -2178,34 +2258,43 @@ static void meta_dc_restore (wmfAPI* API,wmfRecord* Record)
 			
 	clip = (wmfRegion*) P->dc->clip;
 
-	polyrect.TL = (wmfD_Coord*) wmf_malloc (API,clip->numRects * sizeof (wmfD_Coord));
-
-	if (ERR (API))
-	{	WMF_DEBUG (API,"bailing...");
-		return;
-	}
-
-	polyrect.BR = (wmfD_Coord*) wmf_malloc (API,clip->numRects * sizeof (wmfD_Coord));
-
-	if (ERR (API))
-	{	WMF_DEBUG (API,"bailing...");
-		return;
-	}
-
-	polyrect.count = clip->numRects;
-	for (i = 0; i < polyrect.count; i++)
-	{	polyrect.TL[i] = clip->rects[i].TL;
-		polyrect.BR[i] = clip->rects[i].BR;
-	}
-
 	polyrect.dc = P->dc;
+
 	polyrect.width  = 0;
 	polyrect.height = 0;
 
-	if (FR->region_clip) FR->region_clip (API,&polyrect);
+	if (clip->numRects)
+	{	polyrect.TL = (wmfD_Coord*) wmf_malloc (API,clip->numRects * sizeof (wmfD_Coord));
 
-	wmf_free (API,polyrect.TL);
-	wmf_free (API,polyrect.BR);
+		if (ERR (API))
+		{	WMF_DEBUG (API,"bailing...");
+			return;
+		}
+
+		polyrect.BR = (wmfD_Coord*) wmf_malloc (API,clip->numRects * sizeof (wmfD_Coord));
+
+		if (ERR (API))
+		{	WMF_DEBUG (API,"bailing...");
+			return;
+		}
+
+		polyrect.count = clip->numRects;
+		for (i = 0; i < polyrect.count; i++)
+		{	polyrect.TL[i] = clip->rects[i].TL;
+			polyrect.BR[i] = clip->rects[i].BR;
+		}
+
+		wmf_free (API,polyrect.TL);
+		wmf_free (API,polyrect.BR);
+	}
+	else
+	{	polyrect.TL = 0;
+		polyrect.BR = 0;
+
+		polyrect.count = 0;
+	}
+
+	if (FR->region_clip) FR->region_clip (API,&polyrect);
 }
 
 static void meta_text (wmfAPI* API,wmfRecord* Record)
@@ -2223,6 +2312,8 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 
 	wmfDrawText_t drawtext;
 
+	wmfFont* font = 0;
+
 	U16 par_U16;
 	U16 par_U16_x;
 	U16 par_U16_y;
@@ -2256,7 +2347,7 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 			fprintf (stderr,"\t#par=%lu; max. index = 0",Record->size);
 		}
 
-		if (P->dc->textalign & TA_UPDATECP)
+		if (WMF_DC_TEXTALIGN (P->dc) & TA_UPDATECP)
 		{	if ((Record->size) < (1 + (length + 1) / 2))
 			{	WMF_ERROR (API,"Record is too short!");
 				API->err = wmf_E_BadFormat;
@@ -2302,7 +2393,7 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 		{	fprintf (stderr,"\t[0x%04x]",Record->function);
 		}
 
-		if (P->dc->textalign & TA_UPDATECP)
+		if (WMF_DC_TEXTALIGN (P->dc) & TA_UPDATECP)
 		{	if (SCAN (API) && DIAG (API))
 			{	fprintf (stderr,"\t#par=%lu; index 0-1 ignored; max. index = 3",Record->size);
 			}
@@ -2383,8 +2474,10 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 		return;
 	}
 
-	drawtext.font_height = (double) P->dc->font->lfHeight * ABS (P->dc->pixel_height);
-	drawtext.font_ratio = (double) P->dc->font->lfWidth / (double) P->dc->font->lfHeight;
+	font = WMF_DC_FONT (P->dc);
+
+	drawtext.font_height = (double) WMF_FONT_HEIGHT (font) * ABS (P->dc->pixel_height);
+	drawtext.font_ratio = (double) WMF_FONT_WIDTH (font) / (double) WMF_FONT_HEIGHT (font);
 
 	par_U16 = 0;
 	for (i = 0; i < length; i++)
@@ -2398,7 +2491,7 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 	}
 	drawtext.str[length] = '\0';
 
-	width = wmf_ipa_font_stringwidth (API,P->dc->font,drawtext.str);
+	width = wmf_ipa_font_stringwidth (API,font,drawtext.str);
 
 	width = (float) ((double) width * drawtext.font_height * drawtext.font_ratio);
 
@@ -2426,7 +2519,7 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 		width = d_pt.x - t_pt.x;
 	}
 
-	theta = - WMF_TEXT_ANGLE (P->dc->font);
+	theta = - WMF_TEXT_ANGLE (font);
 
 	cos_theta = (float) cos (theta);
 	sin_theta = (float) sin (theta);
@@ -2434,17 +2527,17 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 	/* Okay, have text reference point, string width & font height, & angle (in radians)
 	 * Compute bounding box and adjust for alignment:
 	 */
-	if (P->dc->textalign & TA_BASELINE)
+	if (WMF_DC_TEXTALIGN (P->dc) & TA_BASELINE)
 	{	/* */
 	}
-	else if (P->dc->textalign & TA_BOTTOM)
+	else if (WMF_DC_TEXTALIGN (P->dc) & TA_BOTTOM)
 	{	d_pt.y = - drawtext.font_height / 3; /* This is only approximate... */
 		t_pt.x = - d_pt.y * sin_theta;
 		t_pt.y =   d_pt.y * cos_theta;
 		drawtext.pt.x += t_pt.x;
 		drawtext.pt.y += t_pt.y;
 	}
-	else /* if (P->dc->textalign & TA_TOP) */
+	else /* if (WMF_DC_TEXTALIGN (P->dc) & TA_TOP) */
 	{	d_pt.y = drawtext.font_height;       /* This is only approximate... */
 		t_pt.x = - d_pt.y * sin_theta;
 		t_pt.y =   d_pt.y * cos_theta;
@@ -2452,7 +2545,7 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 		drawtext.pt.y += t_pt.y;
 	}
 
-	if (P->dc->textalign & TA_CENTER)
+	if (WMF_DC_TEXTALIGN (P->dc) & TA_CENTER)
 	{	d_pt.x = width / 2;
 		d_pt.y = - drawtext.font_height * 0.77;
 		t_pt.x = d_pt.x * cos_theta - d_pt.y * sin_theta;
@@ -2483,7 +2576,7 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 		drawtext.pt.x += t_pt.x;
 		drawtext.pt.y += t_pt.y;
 	}
-	else if (P->dc->textalign & TA_RIGHT)
+	else if (WMF_DC_TEXTALIGN (P->dc) & TA_RIGHT)
 	{	d_pt.y = - drawtext.font_height * 0.77;
 		t_pt.x = - d_pt.y * sin_theta;
 		t_pt.y =   d_pt.y * cos_theta;
@@ -2522,11 +2615,11 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 
 		drawtext.pt = d_pt;
 
-		if (P->dc->textalign & TA_UPDATECP)
+		if (WMF_DC_TEXTALIGN (P->dc) & TA_UPDATECP)
 		{	P->current = wmf_L_Coord_translate (API,drawtext.pt);
 		}
 	}
-	else /* if (P->dc->textalign & TA_LEFT) */
+	else /* if (WMF_DC_TEXTALIGN (P->dc) & TA_LEFT) */
 	{	d_pt.y = - drawtext.font_height * 0.77;
 		t_pt.x = - d_pt.y * sin_theta;
 		t_pt.y =   d_pt.y * cos_theta;
@@ -2557,7 +2650,7 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 		drawtext.bbox.BR.y = drawtext.pt.y + t_pt.y;
 		D_Coord_Register (API,drawtext.bbox.BR,0);
 
-		if (P->dc->textalign & TA_UPDATECP)
+		if (WMF_DC_TEXTALIGN (P->dc) & TA_UPDATECP)
 		{	d_pt.x = width;
 			t_pt.x = d_pt.x * cos_theta;
 			t_pt.y = d_pt.x * sin_theta;
@@ -2603,7 +2696,7 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 			d_pt = wmf_D_Coord_translate (API,l_pt);
 			d_pt.x -= o_pt.x;
 			d_pt.y -= o_pt.y;
-			width = wmf_ipa_font_stringwidth (API,P->dc->font,drawtext.str);
+			width = wmf_ipa_font_stringwidth (API,font,drawtext.str);
 			width = (float) ((double) width * drawtext.font_height * ratio);
 			if (d_pt.x < width)
 			{	drawtext.font_ratio = ratio * (d_pt.x / width);
@@ -2637,16 +2730,17 @@ static void meta_text (wmfAPI* API,wmfRecord* Record)
 static void meta_pen_create (wmfAPI* API,wmfRecord* Record) /* complete, I think */
 {	wmfPlayer_t* P  = (wmfPlayer_t*) API->player_data;
 
-	wmfObject* objects;
-	wmfObject* obj_pen;
+	wmfObject* objects = 0;
+	wmfObject* obj_pen = 0;
 
 	wmfRGB color;
 
-	wmfPen* pen;
+	wmfPen* pen = 0;
 
 	U16 par_U16_rg;
 	U16 par_U16_b;
 	U16 par_U16_w;
+	U16 par_U16_s;
 
 	U16 oid_pen;
 
@@ -2673,40 +2767,49 @@ static void meta_pen_create (wmfAPI* API,wmfRecord* Record) /* complete, I think
 	if (SCAN (API) && DIAG (API))
 	{	fprintf (stderr,"\t[0x%04x]",Record->function);
 		fprintf (stderr,"\t#par=%lu; index 2 skipped; max. index = 4",Record->size);
+
+		diagnose_object (API,(unsigned int) oid_pen,obj_pen);
 	}
 
-	pen->lopnStyle = ParU16 (API,Record,0);
+	par_U16_s = ParU16 (API,Record,0);
+
+	WMF_PEN_SET_STYLE  (pen,par_U16_s); /* Need to do all of these separately... */
+	WMF_PEN_SET_ENDCAP (pen,par_U16_s);
+	WMF_PEN_SET_JOIN   (pen,par_U16_s);
+	WMF_PEN_SET_TYPE   (pen,par_U16_s);
 
 	par_U16_w = ParU16 (API,Record,1);
 
 	par_U16_w = MAX (1,par_U16_w);
 
-	pen->width  = (double) par_U16_w * ABS (P->dc->pixel_width );
-	pen->height = (double) par_U16_w * ABS (P->dc->pixel_height);
+	WMF_PEN_SET_WIDTH  (pen,(double) par_U16_w * ABS (P->dc->pixel_width ));
+	WMF_PEN_SET_HEIGHT (pen,(double) par_U16_w * ABS (P->dc->pixel_height));
 
 	par_U16_b  = ParU16 (API,Record,4);
 	par_U16_rg = ParU16 (API,Record,3);
 
 	color = rgb (API,par_U16_rg,par_U16_b);
 
-	pen->lopnColor = color;
+	WMF_PEN_SET_COLOR (pen,&color);
 
 	if (SCAN (API)) wmf_ipa_color_add (API,&color);
 
-	if (pen->lopnStyle != PS_NULL) WMF_DEBUG (API,"Non-null pen style...");
+	if (WMF_PEN_STYLE (pen) != PS_NULL) WMF_DEBUG (API,"Non-null pen style...");
 
-	P->dc->pen = pen;
+	WMF_DC_SET_PEN (P->dc,pen);
 }
 
 static void meta_brush_create (wmfAPI* API,wmfRecord* Record) /* unfinished */
 {	wmfPlayer_t* P  = (wmfPlayer_t*) API->player_data;
 
-	wmfObject* objects;
-	wmfObject* obj_brush;
+	wmfObject* objects = 0;
+	wmfObject* obj_brush = 0;
 
 	wmfRGB color;
 
-	wmfBrush* brush;
+	wmfBrush* brush = 0;
+
+	wmfBMP bmp;
 
 	U16 par_U16_rg;
 	U16 par_U16_b;
@@ -2736,24 +2839,31 @@ static void meta_brush_create (wmfAPI* API,wmfRecord* Record) /* unfinished */
 	if (SCAN (API) && DIAG (API))
 	{	fprintf (stderr,"\t[0x%04x]",Record->function);
 		fprintf (stderr,"\t#par=%lu; max. index = 3",Record->size);
+
+		diagnose_object (API,(unsigned int) oid_brush,obj_brush);
 	}
 
-	brush->lbStyle = ParU16 (API,Record,0);
+	WMF_BRUSH_SET_STYLE (brush,ParU16 (API,Record,0));
 
 	par_U16_b  = ParU16 (API,Record,2);
 	par_U16_rg = ParU16 (API,Record,1);
 
 	color = rgb (API,par_U16_rg,par_U16_b);
 
-	brush->lbColor = color;
+	WMF_BRUSH_SET_COLOR (brush,&color);
 
-	brush->lbHatch  = ParU16 (API,Record,3);
+	WMF_BRUSH_SET_HATCH (brush,ParU16 (API,Record,3));
 
-	brush->bmp.data = 0;
+	bmp.width  = 0;
+	bmp.height = 0;
+
+	bmp.data = 0;
+
+	WMF_BRUSH_SET_BITMAP (brush,&bmp);
 
 	if (SCAN (API)) wmf_ipa_color_add (API,&color);
 
-	P->dc->brush = brush;
+	WMF_DC_SET_BRUSH (P->dc,brush);
 }
 
 static void meta_font_create (wmfAPI* API,wmfRecord* Record)
@@ -2762,10 +2872,10 @@ static void meta_font_create (wmfAPI* API,wmfRecord* Record)
 
 	wmfRecord name_record;
 
-	wmfObject* objects;
-	wmfObject* obj_font;
+	wmfObject* objects = 0;
+	wmfObject* obj_font = 0;
 
-	wmfFont* font;
+	wmfFont* font = 0;
 
 	U16 oid_font;
 
@@ -2777,6 +2887,8 @@ static void meta_font_create (wmfAPI* API,wmfRecord* Record)
 	unsigned long i;
 	unsigned long length;
 
+	char* fontname = 0;
+
 	objects = P->objects;
 
 	i = 0;
@@ -2798,43 +2910,50 @@ static void meta_font_create (wmfAPI* API,wmfRecord* Record)
 	if (SCAN (API) && DIAG (API))
 	{	fprintf (stderr,"\t[0x%04x]",Record->function);
 		fprintf (stderr,"\t#par=%lu; max. index = 8",Record->size);
+
+		diagnose_object (API,(unsigned int) oid_font,obj_font);
 	}
 
 	par_S16_w = ParS16 (API,Record,1);
 	par_S16_h = ParS16 (API,Record,0);
 
-	font->lfHeight         =  ABS (par_S16_h);
-	font->lfWidth          =  ABS (par_S16_w);
+	WMF_FONT_SET_HEIGHT (font,ABS (par_S16_h));
+	WMF_FONT_SET_WIDTH  (font,ABS (par_S16_w));
 
-	font->lfEscapement     =  ParU16 (API,Record,2);
-	font->lfOrientation    =  ParU16 (API,Record,3);
-	font->lfWeight         =  ParU16 (API,Record,4);
+	WMF_FONT_SET_ESCAPEMENT  (font,ParS16 (API,Record,2)); /* text angle */
+	WMF_FONT_SET_ORIENTATION (font,ParS16 (API,Record,3));
+
+	WMF_FONT_SET_WEIGHT (font,ParU16 (API,Record,4));
 
 	par_U16 = ParU16 (API,Record,5);
-	font->lfItalic         =  par_U16       & 0xff;
-	font->lfUnderline      = (par_U16 >> 8) & 0xff;
+	WMF_FONT_SET_ITALIC (font,par_U16 & 0xff);
+
+	WMF_TEXT_SET_UNDERLINE (font,(par_U16 >> 8) & 0xff);
 
 	par_U16 = ParU16 (API,Record,6);
-	font->lfStrikeOut      =  par_U16       & 0xff;
-	font->lfCharSet        = (par_U16 >> 8) & 0xff;
+	WMF_TEXT_SET_STRIKEOUT (font,par_U16 & 0xff);
+
+	WMF_FONT_SET_CHARSET (font,(par_U16 >> 8) & 0xff);
 
 	par_U16 = ParU16 (API,Record,7);
-	font->lfOutPrecision   =  par_U16       & 0xff;
-	font->lfClipPrecision  = (par_U16 >> 8) & 0xff;
+	WMF_FONT_SET_OUT (font,par_U16 & 0xff);
+
+	WMF_FONT_SET_CLIP (font,(par_U16 >> 8) & 0xff);
 
 	par_U16 = ParU16 (API,Record,8);
-	font->lfQuality        =  par_U16       & 0xff;
-	font->lfPitchAndFamily = (par_U16 >> 8) & 0xff;
+	WMF_FONT_SET_QUALITY (font,par_U16 & 0xff);
 
-	if (font->lfWidth == 0)
-	{	font->lfWidth = font->lfHeight;
+	WMF_FONT_SET_PITCH (font,(par_U16 >> 8) & 0xff);
+
+	if (WMF_FONT_WIDTH (font) == 0)
+	{	WMF_FONT_SET_WIDTH (font,WMF_FONT_HEIGHT (font));
 	}
 
 	name_record = OffsetRecord (API,Record,9);
 
 	length = (name_record.size) * 2;
 
-	font->lfFaceName = (char*) wmf_malloc (API,length + 1);
+	fontname = (char*) wmf_malloc (API,length + 1);
 
 	if (ERR (API))
 	{	WMF_DEBUG (API,"bailing...");
@@ -2844,14 +2963,16 @@ static void meta_font_create (wmfAPI* API,wmfRecord* Record)
 	par_U16 = 0;
 	for (i = 0; i < length; i++)
 	{	if (i & 1)
-		{	font->lfFaceName[i] =  (par_U16 >> 8) & 0xff;
+		{	fontname[i] = (par_U16 >> 8) & 0xff;
 		}
 		else
 		{	par_U16 = ParU16 (API,&name_record,i/2);
-			font->lfFaceName[i] =  par_U16 & 0xff;
+			fontname[i] = par_U16 & 0xff;
 		}
 	}
-	font->lfFaceName[length] = '\0';
+	fontname[length] = '\0';
+
+	WMF_FONT_SET_NAME (font,fontname); /* must not free fontname */
 
 	FD->map (API,font); /* Determine/Load freetype face */
 
@@ -2860,7 +2981,28 @@ static void meta_font_create (wmfAPI* API,wmfRecord* Record)
 		return;
 	}
 
-	P->dc->font = font;
+	WMF_DC_SET_FONT (P->dc,font);
+}
+
+static void meta_palette_create (wmfAPI* API,wmfRecord* Record)
+{	wmfPlayer_t* P  = (wmfPlayer_t*) API->player_data;
+
+	wmfObject* objects = 0;
+
+	unsigned long i;
+
+	objects = P->objects;
+
+	i = 0;
+	while (objects[i].type && (i < NUM_OBJECTS (API))) i++;
+
+	if (i == NUM_OBJECTS (API))
+	{	WMF_ERROR (API,"Object out of range!");
+		API->err = wmf_E_BadFormat;
+		return;
+	}
+
+	objects[i].type = OBJ_PAL;
 }
 
 static void meta_delete (wmfAPI* API,wmfRecord* Record)
@@ -2889,6 +3031,10 @@ static void meta_delete (wmfAPI* API,wmfRecord* Record)
 
 	obj = objects + oid;
 
+	if (SCAN (API) && DIAG (API))
+	{	diagnose_object (API,(unsigned int) oid,obj);
+	}
+
 	if (obj->type == OBJ_BRUSH)
 	{	if (obj->obj.brush.lbStyle == BS_DIBPATTERN)
 		{	if (PLAY (API) && FR->bmp_free) FR->bmp_free (API,&(obj->obj.brush.bmp));
@@ -2947,4 +3093,35 @@ recommended approach in today's font world.
 	/* META_ESCAPE: ?? */
 }
 
+static void diagnose_object (wmfAPI* API,unsigned int oid,wmfObject* obj)
+{	fprintf (stderr,"\toid=%u / %u ",(unsigned) oid,(unsigned) NUM_OBJECTS (API));
+
+	switch (obj->type)
+	{
+	case OBJ_BRUSH:
+		fprintf (stderr,"(brush)");
+	break;
+
+	case OBJ_PEN:
+		fprintf (stderr,"(pen)");
+	break;
+
+	case OBJ_FONT:
+		fprintf (stderr,"(font)");
+	break;
+
+	case OBJ_REGION:
+		fprintf (stderr,"(region)");
+	break;
+
+	case OBJ_PAL:
+		fprintf (stderr,"(palette)");
+	break;
+
+	default:
+		fprintf (stderr,"(other [%u])",(unsigned) obj->type);
+	break;
+	}
+}
+
 #endif /* ! WMFPLAYER_META_H */
